from __future__ import annotations

from functools import cache, partial
from typing import TYPE_CHECKING, Callable

import graphene
from graphene.types.objecttype import ObjectTypeOptions

if TYPE_CHECKING:
    from django.db import models

    from graphene_django_extensions.typing import Any, GQLInfo, Sequence

__all__ = [
    "LanguageListField",
    "LanguageString",
    "Translatable",
    "TranslationsField",
    "add_translatable_fields",
    "get_available_languages",
    "get_translatable_fields",
]


def get_translatable_fields(model: type[models.Model]) -> list[str]:
    """If `django-modeltranslation` is installed, find all translatable fields in the given model."""
    try:
        from modeltranslation.manager import get_translatable_fields_for_model
    except ImportError:
        return []

    return get_translatable_fields_for_model(model) or []


def add_translatable_fields(model: type[models.Model], fields: Sequence[str]) -> Sequence[str]:
    """
    If `django-modeltranslation` is installed, find and add all translation fields
    to the given fields list, for the given fields, in the given model.
    """
    try:
        from modeltranslation.manager import get_translatable_fields_for_model
        from modeltranslation.utils import get_translation_fields
    except ImportError:
        return fields

    translatable_fields: list[str] = get_translatable_fields_for_model(model) or []
    new_fields: list[str] = []
    for field in fields:
        new_fields.append(field)
        if field not in translatable_fields:
            continue
        fields = get_translation_fields(field)
        new_fields.extend(fields)

    return new_fields


@cache
def get_available_languages() -> list[str]:
    """Get a list of all available translation languages."""
    try:
        from modeltranslation.settings import AVAILABLE_LANGUAGES
    except ImportError:
        return []

    return AVAILABLE_LANGUAGES


class BaseTranslatable(graphene.ObjectType):
    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        # This method is run only on subclasses, which is why we need this base class.
        _meta = ObjectTypeOptions(cls)
        _meta.fields = {
            lang: graphene.Field(
                graphene.String,
                description=f"Translation in {lang}.",
            )
            for lang in get_available_languages()
        }
        super().__init_subclass_with_meta__(_meta=_meta, **options)


class Translatable(BaseTranslatable): ...


class LanguageString(graphene.ObjectType):
    language = graphene.String(required=True, description="Language code.")
    value = graphene.String(required=False, description="Translation in the given language.")


class TranslationsField(graphene.Field):
    def __init__(self, field_name: str, **kwargs: Any) -> None:
        kwargs["type_"] = graphene.NonNull(Translatable)
        self.field_name = field_name
        self.languages = get_available_languages()
        super().__init__(**kwargs)

    def wrap_resolve(self, parent_resolver: Callable) -> Callable:
        if isinstance(parent_resolver, partial):  # shorthand for default resolver
            return self.field_resolver
        return parent_resolver  # pragma: no cover

    def field_resolver(self, root: models.Model, info: GQLInfo) -> dict[str, str]:
        return {lang: getattr(root, f"{self.field_name}_{lang}") for lang in self.languages}


class LanguageListField(graphene.Field):
    def __init__(self, field_name: str, **kwargs: Any) -> None:
        kwargs["type_"] = graphene.List(graphene.NonNull(LanguageString))
        self.field_name = field_name
        self.languages = get_available_languages()
        super().__init__(**kwargs)

    def wrap_resolve(self, parent_resolver: Callable) -> Callable:
        if isinstance(parent_resolver, partial):  # shorthand for default resolver
            return self.field_resolver
        return parent_resolver  # pragma: no cover

    def field_resolver(self, root: models.Model, info: GQLInfo) -> list[dict[str, str]]:
        return [{"language": lang, "value": getattr(root, f"{self.field_name}_{lang}")} for lang in self.languages]



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from query_optimizer.filter_info import get_filter_info

from .constants import Operation
from .settings import gdx_settings
from .typing import StrEnum

if TYPE_CHECKING:
    from .typing import Any

__all__ = [
    "get_filter_info",
    "get_nested",
    "get_operator_enum",
]


def get_nested(obj: dict | list | None, /, *args: str | int, default: Any = None) -> Any:
    """
    Get value from a nested structure containing dicts with string keys or lists,
    where the keys and list indices might not exist.

    1) `data["foo"][0]["bar"]["baz"]`
     - Might raise a `KeyError` or `IndexError` if any of the keys or indices don't exist.

    2) `get_nested(data, "foo", 0, "bar", "baz")`
     - Will return `None` (default) if any of the keys or indices don't exist.
    """
    if not args:
        return obj if obj is not None else default

    arg, args = args[0], args[1:]

    if isinstance(arg, int):
        obj = obj or []
        try:
            obj = obj[arg]
        except (IndexError, KeyError):
            obj = None
        return get_nested(obj, *args, default=default)

    obj = obj or {}
    try:
        return get_nested(obj.get(arg), *args, default=default)
    except AttributeError:
        obj = None
        return get_nested(obj, *args, default=default)


@cache
def get_operator_enum() -> type[Operation]:
    if not gdx_settings.EXTEND_USER_DEFINED_FILTER_OPERATIONS:  # pragma: no cover
        return Operation

    current_members: dict[str, str] = {key: value.value for key, value in Operation._member_map_.items()}
    for operator in gdx_settings.EXTEND_USER_DEFINED_FILTER_OPERATIONS:
        current_members[operator.upper()] = operator.upper()

    return StrEnum(Operation.__name__, current_members)  # type: ignore[return-value]
